# Generated by Django 4.2.22 on 2025-06-30 17:35

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('hub', '0005_add_task_canceller'),
    ]

    operations = [
        migrations.AlterField(
            model_name='task',
            name='canceled_by',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='canceled_by1', to=settings.AUTH_USER_MODEL),
        ),
    ]
